from . import wrappers
from .wrapper import (
    GRPCInvisibleSequenceWrapper,
    GRPCInvisibleWrapper,
    GRPCMessageWrapper,
    GRPCRepeatedMessageWrapper,
)

try:
    from ._version import __version__, __version_tuple__
except ImportError:  # pragma: no cover
    __version__ = version = "unknown"
    __version_tuple__ = version_tuple = ("unknown",)  # type: ignore

__author__ = "Matthew Wardrop"
__author_email__ = "mpwardrop@gmail.com"

__all__ = [
    # "__version__",
    # "__version_tuple__",
    # "__author__",
    # "__author_email__",
    "wrappers",
    "GRPCInvisibleSequenceWrapper",
    "GRPCInvisibleWrapper",
    "GRPCMessageWrapper",
    "GRPCRepeatedMessageWrapper",
]



import collections
import inspect
import logging
import warnings
from abc import abstractmethod
from enum import Enum

import google.protobuf.descriptor
import google.protobuf.message
import google.protobuf.internal.containers
from google.protobuf.symbol_database import Default as get_grpc_symbol_database
from interface_meta import InterfaceMeta, override

from .utils import cast_enum_type_to_int


GRPC_SYMBOL_DATABASE = get_grpc_symbol_database()


class GRPCMessageWrapperMeta(InterfaceMeta):

    _WRAPPERS = {}
    _ENUM_TYPES = {}

    def __register_implementation__(cls):
        if not cls._MESSAGE_TYPE:
            return

        if not inspect.isclass(cls._MESSAGE_TYPE) or not issubclass(
            cls._MESSAGE_TYPE, google.protobuf.message.Message
        ):
            raise ValueError(
                f"Class `{cls.__name__}` has an invalid `_MESSAGE_TYPE`. Object of type `{cls._MESSAGE_TYPE.__class__.__name__}` is not a subclass of `google.protobuf.message.Message`."
            )

        # Register subclass
        cls._WRAPPERS[cls._MESSAGE_TYPE.DESCRIPTOR.full_name] = cls

        # Add message fields as attributes of the class
        for field, desc in cls._MESSAGE_TYPE.DESCRIPTOR.fields_by_name.items():

            if field in cls._IGNORED_FIELDS:
                continue

            if hasattr(cls, field):
                # Check that it is not a collision at the metaclass level
                collides_in_metaclass = True
                for base in cls.mro():
                    if field in base.__dict__:
                        collides_in_metaclass = False
                        break

                if not collides_in_metaclass:
                    logging.warning(
                        f"Field `{field}` of `{cls._MESSAGE_TYPE}` is masked by class attributes."
                    )
                    continue

            # Add field wrappers so that the attributes appear in class documentation
            setattr(cls, field, cls.__get_field_property_method(field))

            # Add enum specifications to class definition
            if desc.enum_type:
                if desc.enum_type.full_name not in cls._ENUM_TYPES:
                    cls._ENUM_TYPES[desc.enum_type.full_name] = Enum(
                        desc.enum_type.name,
                        [(v.name, v.number) for v in desc.enum_type.values],
                    )

    def __get_field_property_method(cls, field):
        def get_field(self):
            return self.__getattr__(field)

        get_field.__name__ = field

        def set_field(self):
            return self.__setattr__(field)

        return property(fget=get_field, fset=set_field)

    def for_kind(cls, kind):
        if kind is None:
            return
        if isinstance(kind, google.protobuf.descriptor.Descriptor):
            kind = kind.full_name
        if isinstance(kind, str):
            kind = GRPC_SYMBOL_DATABASE.GetSymbol(kind)
        assert inspect.isclass(kind) and issubclass(
            kind, google.protobuf.message.Message
        )
        if kind.DESCRIPTOR.full_name not in cls._WRAPPERS:
            GRPCMessageWrapperMeta(
                kind.DESCRIPTOR.full_name,
                (GRPCMessageWrapper,),
                {"_MESSAGE_TYPE": kind},
            )
        return cls._WRAPPERS[kind.DESCRIPTOR.full_name]

    def for_message(cls, message):
        if isinstance(message, google.protobuf.message.Message):
            return cls.for_kind(type(message))(message)
        return message

    def from_json(cls, json, message_type=None):
        message_type = message_type or getattr(cls, "_MESSAGE_TYPE", None)
        if not message_type:
            raise RuntimeError(
                "`from_json()` can only be called on a wrapper which has `_MESSAGE_TYPE` set, or by passing in message_type specifically."
            )
        from google.protobuf.json_format import ParseDict

        return cls.for_message(ParseDict(json, message_type()))


class GRPCMessageWrapper(metaclass=GRPCMessageWrapperMeta):

    _MESSAGE_TYPE = None
    _IMMUTABLE_FIELDS = ()
    _IGNORED_FIELDS = ()

    __slots__ = ["__message", "__persisted_fields"]

    def __init__(self, _message=None, **kwargs):
        if self._MESSAGE_TYPE is None:
            raise RuntimeError(
                "`GRPCMessageWrapper` should not be instantiated directly. "
                "Please use `GRPCMessageWrapper.for_message(...)` instead."
            )
        self.__persisted_fields = {}
        if isinstance(_message, GRPCMessageWrapper):
            raise ValueError(
                "Cannot wrap already wrapped message. Use `.for_message(...)` instead."
            )
        self.__message = _message or self._MESSAGE_TYPE()
        self._init(**kwargs)

    def _init(self, **kwargs):
        self << kwargs

    @property
    def _message(self):
        return self.__message

    @property
    def _message_copy(self):
        message = type(self.__message)()
        message.CopyFrom(self.__message)
        return message

    def __dir__(self):
        return set(
            [
                *(
                    f.name
                    for f in self.__message.DESCRIPTOR.fields
                    if f.name not in self._IGNORED_FIELDS
                ),
                *super().__dir__(),
            ]
        )

    @property
    def __message_fields(self):
        return set(self._message.DESCRIPTOR.fields_by_name)

    def __get_field_descriptor(self, field):
        return self._message.DESCRIPTOR.fields_by_name[field]

    def __get_wrapped_value(self, field, value, evaluate_getters=True):
        descriptor = self.__get_field_descriptor(field)

        # Warn about deprecated field usage
        if descriptor.GetOptions().deprecated:
            warnings.warn(
                f"`{self._message.__class__.__name__}.{field}` is deprecated and may disappear in the future.",
                DeprecationWarning,
            )

        if descriptor.label == descriptor.LABEL_REPEATED:
            if (
                descriptor.message_type
                and descriptor.message_type.GetOptions().map_entry
            ):
                wrapper = GRPCMessageWrapper.for_kind(
                    descriptor.message_type.fields_by_name["value"].message_type
                )
                value = GRPCMapMessageWrapper(value, wrapper=wrapper)
            else:
                wrapper = (
                    GRPCMessageWrapper._ENUM_TYPES[descriptor.enum_type.full_name]
                    if descriptor.enum_type
                    else GRPCMessageWrapper.for_kind(descriptor.message_type)
                )
                value = GRPCRepeatedMessageWrapper(value, wrapper=wrapper)
        elif descriptor.enum_type:
            value = GRPCMessageWrapper._ENUM_TYPES[descriptor.enum_type.full_name](
                value
            )
        else:
            value = GRPCMessageWrapper.for_message(value)

        if evaluate_getters and hasattr(value, "__get__"):
            value = value.__get__() if self.__message.HasField(field) else None

        return value

    def __get_wrapped_field_value(self, field, evaluate_getters=True):
        return self.__get_wrapped_value(
            field=field,
            value=getattr(self.__message, field),
            evaluate_getters=evaluate_getters,
        )

    def __set_field_value(self, field, value):
        descriptor = self.__get_field_descriptor(field)

        # Allow wrappers to handle setting if this is not directly set on the
        # message wrapped by this instance.
        wrapper = self.__get_wrapped_field_value(field, evaluate_getters=False)
        if hasattr(wrapper, "__set__"):
            return wrapper.__set__(value)

        # Prepare enum values
        if descriptor.enum_type:
            if self._MESSAGE_TYPE:
                value = cast_enum_type_to_int(
                    enum_type=GRPCMessageWrapper._ENUM_TYPES[
                        descriptor.enum_type.full_name
                    ],
                    value=value,
                )
            else:
                logging.warning(
                    f"`{field}` is an enum type, but messages of type `{self._message.__class__}` have not yet been individually wrapped and so only integer enum code will be accepted. Use with care."
                )

        self.__persisted_fields.pop(field, None)
        # If value is wrapped, unwrap it before assigning to the GRPC message.
        if isinstance(value, GRPCMessageWrapper):
            value = value._message
        return self.__message.MergeFrom(self.__message.__class__(**{field: value}))

    # External API

    def __getattr__(self, attr):
        if attr not in self.__message_fields or attr in self._IGNORED_FIELDS:
            raise AttributeError(attr)

        if attr not in self.__persisted_fields:
            self.__persisted_fields[attr] = self.__get_wrapped_field_value(field=attr)
        return self.__persisted_fields[attr]

    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            return super().__setattr__(attr, value)
        elif attr not in self.__message_fields or attr in self._IGNORED_FIELDS:
            raise AttributeError(attr)
        elif attr in self._IMMUTABLE_FIELDS:
            raise AttributeError(f"`{attr}` is immutable.")

        return self.__set_field_value(field=attr, value=value)

    def __delattr__(self, attr):
        self.__message.ClearField(attr)
        self.__persisted_fields.pop(attr, None)

    def __repr__(self):
        return f"GRPCMessageWrapper<{self.__class__.__name__}>"

    def __lshift__(self, other):
        if not isinstance(other, dict):
            return NotImplemented
        for key, value in other.items():
            setattr(self, key, value)
        return self

    def _compare(self, ref=None):
        ref = ref or type(self.__message)()
        if isinstance(ref, GRPCMessageWrapper):
            ref = ref._message
        if self.__message == ref:
            return

        comparison = {}

        for field_name in self.__message_fields:
            if (
                field_name in self._IGNORED_FIELDS
                or field_name in self._IMMUTABLE_FIELDS
            ):
                continue
            if getattr(ref, field_name) != getattr(self.__message, field_name):
                field_value = self.__get_wrapped_field_value(
                    field_name, evaluate_getters=False
                )
                if isinstance(field_value, GRPCMessageWrapper):
                    nested_comparison = field_value._compare(
                        ref=getattr(ref, field_name)
                    )
                    if nested_comparison:
                        comparison[field_name] = nested_comparison
                else:
                    comparison[field_name] = (
                        self.__get_wrapped_value(
                            field=field_name, value=getattr(ref, field_name)
                        ),
                        self.__get_wrapped_field_value(field_name),
                    )

        return comparison

    def _to_json(self):
        from google.protobuf.json_format import MessageToDict

        try:
            return MessageToDict(
                self._message, always_print_fields_with_no_presence=True
            )
        except TypeError:
            return MessageToDict(self._message, including_default_value_fields=True)


class GRPCInvisibleWrapper(GRPCMessageWrapper):
    @abstractmethod
    def __get__(self):
        raise NotImplementedError

    @abstractmethod
    def __set__(self, value):
        raise NotImplementedError


class GRPCInvisibleSequenceWrapper(GRPCInvisibleWrapper):

    _SEQUENCE_ATTR = None

    @override
    def __get__(self):
        return getattr(self, self._SEQUENCE_ATTR)

    @override
    def __set__(self, value):
        sequence = getattr(self, self._SEQUENCE_ATTR)
        while sequence:
            sequence.pop()
        sequence.extend(
            v._message if isinstance(v, GRPCMessageWrapper) else v for v in value
        )

    @override
    def _compare(self, ref=None):
        return super()._compare(ref=ref).get(self._SEQUENCE_ATTR)


class GRPCContainerWrapper:
    def __init__(self, wrapper=None):
        self._wrapper = wrapper
        self.__wrapper_cache = {}

    def _wrap(self, obj):
        cache_key = id(obj)
        if cache_key not in self.__wrapper_cache:
            self.__wrapper_cache[cache_key] = (
                self._wrapper(obj)
                if self._wrapper
                else GRPCMessageWrapper.for_message(obj)
            )
        return self.__wrapper_cache[cache_key]

    def _unwrap(self, obj):
        if isinstance(obj, GRPCMessageWrapper):
            return obj._message
        if isinstance(self._wrapper, Enum):
            return cast_enum_type_to_int(self._wrapper, obj)
        return obj

    def _wrapper_purge(self, obj):
        self.__wrapper_cache.pop(id(obj), None)


class GRPCMapMessageWrapper(GRPCContainerWrapper, collections.abc.MutableMapping):
    def __init__(self, map, wrapper=None):
        self._map = map
        GRPCContainerWrapper.__init__(self, wrapper=wrapper)

    # MutableMapping implementations

    def __getitem__(self, key):
        if key in self._map:
            return self._wrap(self._map[key])
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._wrapper_purge(self._map[key])
        if self._wrapper:
            self._map[key].MergeFrom(self._unwrap(value))
        else:
            self._map.__setitem__(key, value)

    def __delitem__(self, key):
        self._wrapper_purge(self._map[key])
        return self._map.__delitem__(key)

    def __iter__(self):
        return self._map.__iter__()

    def __len__(self):
        return self._map.__len__()

    # Additional convenience methods

    def add(self, key, **kwargs):
        self[key] = self._wrapper(**kwargs)

    def __repr__(self):
        return repr(dict(self))


class GRPCRepeatedMessageWrapper(GRPCContainerWrapper, collections.abc.MutableSequence):
    def __init__(self, sequence, wrapper=None):
        self._sequence = sequence
        GRPCContainerWrapper.__init__(self, wrapper=wrapper)

    def __set__(self, values):
        # To avoid clearing existing values if new values are invalid, we extend
        # first, and then clear up to the first new index.
        index = len(self._sequence)
        self.extend(values)
        for i in range(index):
            self.pop(0)

    # MutableSequence implementations

    def __getitem__(self, index):
        return self._wrap(self._sequence[index])

    def __setitem__(self, index, obj):
        self._wrapper_purge(self._sequence[index])
        return self._sequence.__setitem__(index, self._unwrap(obj))

    def __delitem__(self, index):
        self._wrapper_purge(self._sequence[index])
        return self._sequence.__delitem__(index)

    def __len__(self):
        return len(self._sequence)

    def insert(self, index, obj):
        obj = self._unwrap(obj)
        if isinstance(
            self._sequence,
            google.protobuf.internal.containers.RepeatedCompositeFieldContainer,
        ):
            if index == len(self):
                self._sequence.extend([obj])
            else:
                raise RuntimeError(
                    "Insertion is only supported for repeated scalar containers."
                )
        else:
            self._sequence.insert(index, obj)

    # Additional convenience methods

    def add(self, **kwargs):
        if isinstance(
            self._sequence,
            google.protobuf.internal.containers.RepeatedCompositeFieldContainer,
        ):
            self.append(self._wrapper(**kwargs))
        else:
            RuntimeError(
                "Addition is only supported for repeated composite containers. Use `.append()` for scalar values."
            )

    def __repr__(self):
        return repr(list(self))



import pytest


@pytest.fixture
def message():
    from example_grpc.example_pb2 import ExampleMessage

    return ExampleMessage()


@pytest.fixture
def wrapped(message):
    from grpc_wrappers import GRPCMessageWrapper

    return GRPCMessageWrapper.for_message(message)



import arrow
import google.protobuf.timestamp_pb2
import google.protobuf.wrappers_pb2
from interface_meta import override

from ..wrapper import GRPCInvisibleWrapper


class ProtobufWrapper(GRPCInvisibleWrapper):
    @override
    def __get__(self):
        return self._message.value

    @override
    def __set__(self, value):
        self._message.value = value

    @override
    def _compare(self, ref=None):
        return super()._compare(ref=ref).get("value")


class ProtobufDoubleWrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.DoubleValue


class ProtobufFloatWrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.FloatValue


class ProtobufInt64Wrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.Int64Value


class ProtobufUInt64Wrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.UInt64Value


class ProtobufInt32Wrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.Int32Value


class ProtobufUInt32Wrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.UInt32Value


class ProtobufBytesWrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.BytesValue


class ProtobufStringWrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.StringValue


class ProtobufBoolWrapper(ProtobufWrapper):

    _MESSAGE_TYPE = google.protobuf.wrappers_pb2.BoolValue


class ProtobufTimestampWrapper(GRPCInvisibleWrapper):

    _MESSAGE_TYPE = google.protobuf.timestamp_pb2.Timestamp

    @override
    def __get__(self):
        return (
            self._message.ToDatetime()
            if self._message.seconds > 0 or self._message.nanos > 0
            else None
        )

    @override
    def __set__(self, value):
        self._message.FromDatetime(arrow.get(value))

    @override
    def _compare(self, ref=None):
        if self._message.seconds != ref.seconds:
            return (ref.ToDatetime(), self._message.ToDatetime())



def cast_enum_type_to_int(enum_type, value):
    if isinstance(value, str):
        value = enum_type[value]
    elif isinstance(value, int):
        value = enum_type(value)
    if not isinstance(value, enum_type):
        raise ValueError(value)
    return value.value



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: example.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rexample.proto\x12\x13grpc_wrappers.tests\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/wrappers.proto"\xbe\x05\n\x0e\x45xampleMessage\x12\x12\n\nbool_value\x18\x01 \x01(\x08\x12\x11\n\tint_value\x18\x02 \x01(\x03\x12\x14\n\x0cstring_value\x18\x03 \x01(\t\x12\x34\n\nenum_value\x18\x04 \x01(\x0e\x32 .grpc_wrappers.tests.ExampleEnum\x12\x37\n\nsubmessage\x18\x05 \x01(\x0b\x32#.grpc_wrappers.tests.ExampleMessage\x12#\n\x17\x64\x65precated_string_value\x18\x06 \x01(\tB\x02\x18\x01\x12\x12\n\nint_values\x18\x07 \x03(\x03\x12\x35\n\x0b\x65num_values\x18\x08 \x03(\x0e\x32 .grpc_wrappers.tests.ExampleEnum\x12\x38\n\x0bsubmessages\x18\t \x03(\x0b\x32#.grpc_wrappers.tests.ExampleMessage\x12L\n\x11nested_submessage\x18\n \x01(\x0b\x32\x31.grpc_wrappers.tests.ExampleMessage.NestedMessage\x12\x36\n\x12wrapped_bool_value\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x38\n\x13wrapped_float_value\x18\x0c \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12:\n\x14wrapped_string_value\x18\r \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x35\n\x11wrapped_timestamp\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a#\n\rNestedMessage\x12\x12\n\nbool_value\x18\x01 \x01(\x08*0\n\x0b\x45xampleEnum\x12\x05\n\x01\x41\x10\x00\x12\x05\n\x01\x42\x10\x01\x12\x05\n\x01\x43\x10\x02\x12\x05\n\x01\x44\x10\x03\x12\x05\n\x01\x45\x10\x04\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "example_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _EXAMPLEMESSAGE.fields_by_name["deprecated_string_value"]._options = None
    _EXAMPLEMESSAGE.fields_by_name[
        "deprecated_string_value"
    ]._serialized_options = b"\030\001"
    _globals["_EXAMPLEENUM"]._serialized_start = 808
    _globals["_EXAMPLEENUM"]._serialized_end = 856
    _globals["_EXAMPLEMESSAGE"]._serialized_start = 104
    _globals["_EXAMPLEMESSAGE"]._serialized_end = 806
    _globals["_EXAMPLEMESSAGE_NESTEDMESSAGE"]._serialized_start = 771
    _globals["_EXAMPLEMESSAGE_NESTEDMESSAGE"]._serialized_end = 806
# @@protoc_insertion_point(module_scope)



from enum import Enum
from typing import Union

import pytest

from example_grpc.example_pb2 import ExampleMessage
from grpc_wrappers import GRPCMessageWrapper, GRPCRepeatedMessageWrapper


def test_direct_instantiation(message):
    with pytest.raises(RuntimeError):
        GRPCMessageWrapper(message)


def test_basic_types(wrapped):
    assert wrapped.bool_value is False
    assert wrapped.int_value == 0
    assert wrapped.string_value == ""

    wrapped.bool_value = True
    wrapped.int_value = 100
    wrapped.string_value = "Test String"
    assert wrapped.bool_value is True
    assert wrapped.int_value == 100
    assert wrapped.string_value == "Test String"

    with pytest.raises(TypeError):
        wrapped.bool_value = "INVALID"

    with pytest.raises(TypeError):
        wrapped.string_value = 10

    assert set(wrapped._compare()) == {
        "bool_value",
        "int_value",
        "string_value",
    }


def test_enum_types(wrapped):
    assert isinstance(wrapped.enum_value, Enum)
    assert wrapped.enum_value == wrapped.enum_value.A
    assert set(wrapped.enum_value.__class__.__members__) == set("ABCDE")

    wrapped.enum_value = "B"
    assert wrapped.enum_value == wrapped.enum_value.B

    wrapped.enum_value = 2
    assert wrapped.enum_value == wrapped.enum_value.C

    with pytest.raises(KeyError):
        wrapped.enum_value = "INVALID"

    assert set(wrapped._compare()) == {"enum_value"}


def test_submessage(wrapped):
    assert isinstance(wrapped.submessage, GRPCMessageWrapper)

    assert wrapped.submessage.bool_value is False

    wrapped.submessage.bool_value = True
    assert wrapped.submessage.bool_value is True

    changes = wrapped._compare()
    assert set(changes) == {"submessage"}
    assert isinstance(changes["submessage"], dict)
    assert set(changes["submessage"]) == {"bool_value"}


def test_nested_submessage(wrapped):
    assert isinstance(wrapped.nested_submessage, GRPCMessageWrapper)

    assert wrapped.nested_submessage.bool_value is False

    wrapped.nested_submessage.bool_value = True
    assert wrapped.nested_submessage.bool_value is True

    changes = wrapped._compare()
    assert set(changes) == {"nested_submessage"}
    assert isinstance(changes["nested_submessage"], dict)
    assert set(changes["nested_submessage"]) == {"bool_value"}


def test_repeated_scalar_types(wrapped):
    assert isinstance(wrapped.int_values, GRPCRepeatedMessageWrapper)

    assert len(wrapped.int_values) == 0

    wrapped.int_values.append(10)
    assert len(wrapped.int_values) == 1

    wrapped.int_values.extend([11, 12, 13])
    assert len(wrapped.int_values) == 4

    wrapped.int_values = [1, 2, 3, 4]
    assert list(wrapped.int_values) == [1, 2, 3, 4]

    with pytest.raises(TypeError):
        wrapped.int_values = "String"
    assert len(wrapped.int_values) == 4

    assert set(wrapped._compare()) == {"int_values"}


def test_repeated_enum_types(wrapped):
    assert isinstance(wrapped.enum_values, GRPCRepeatedMessageWrapper)

    assert len(wrapped.enum_values) == 0

    wrapped.enum_values.append(1)
    assert len(wrapped.enum_values) == 1

    wrapped.enum_values.extend([2, 3, 4])
    assert len(wrapped.enum_values) == 4

    wrapped.enum_values = [1, 2, 3, 4]
    assert [ev.value for ev in wrapped.enum_values] == [1, 2, 3, 4]

    with pytest.raises(Exception):
        # Could be a TypeError or ValueError depending on protobuf version
        wrapped.enum_values = "String"
    assert len(wrapped.enum_values) == 4

    assert set(wrapped._compare()) == {"enum_values"}


def test_repeated_message_types(wrapped):
    assert isinstance(wrapped.submessages, GRPCRepeatedMessageWrapper)

    assert len(wrapped.submessages) == 0

    wrapped.submessages.append(ExampleMessage())
    assert len(wrapped.submessages) == 1

    wrapped.submessages.extend([ExampleMessage(), ExampleMessage(), ExampleMessage()])
    assert len(wrapped.submessages) == 4

    wrapped.submessages = [ExampleMessage()] * 4
    assert len(wrapped.submessages) == 4

    with pytest.raises(TypeError):
        wrapped.submessages = "String"
    assert len(wrapped.submessages) == 4

    assert set(wrapped._compare()) == {"submessages"}



import time

import arrow
import pytest


def test_google_wrappers_get(wrapped):
    assert wrapped.wrapped_bool_value is None
    assert wrapped.wrapped_float_value is None
    assert wrapped.wrapped_string_value is None
    assert wrapped.wrapped_timestamp is None


def test_google_wrappers_set(wrapped):
    timestamp = time.time()

    wrapped.wrapped_bool_value = True
    wrapped.wrapped_float_value = 1000
    wrapped.wrapped_string_value = "Test String"
    wrapped.wrapped_timestamp = timestamp

    assert wrapped.wrapped_bool_value is True
    assert wrapped.wrapped_float_value == 1000
    assert wrapped.wrapped_string_value == "Test String"
    assert arrow.get(wrapped.wrapped_timestamp).timestamp() == pytest.approx(timestamp)

    assert set(wrapped._compare()) == {
        "wrapped_bool_value",
        "wrapped_float_value",
        "wrapped_string_value",
        "wrapped_timestamp",
    }
